// file      : xsd/cxx/tree/parser-source.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/parser-source.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      struct Element: Traversal::Element,
                      protected virtual Context
      {
        Element (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          String name (e.name ());
          //String string_type (L"::std::basic_string< " + char_type + L" >");
          String properties_type (L"::xsd::cxx::tree::properties< " +
                                  char_type + L" >");

          // Note that I am using fq-name in function calls because g++ gets
          // confused if the name is 'type'. (see tests/schema/anonymous)
          //

          // URI.
          //
          os << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << id (name) << " (" << string_type << " const& u," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << properties_type << " const& p)"
             << "{"
             << "::xsd::cxx::xml::auto_initializer i (" << endl
             << "(f & ::xsd::cxx::tree::flags::dont_initialize) == 0," << endl
             << "(f & ::xsd::cxx::tree::flags::keep_dom) == 0);"
             << endl
             << "::xsd::cxx::tree::error_handler< " << char_type << " > h;"
             << endl
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << "::xsd::cxx::xml::dom::dom< " << char_type << " > (" << endl
             << "u, h, p, (f & ::xsd::cxx::tree::flags::dont_validate) == 0));"
             << endl
             << "h.throw_if_failed< ::xsd::cxx::tree::parsing< " <<
            char_type << " > > ();"
             << endl
             << "return " << fq_name (e) << " (" << endl
             << "static_cast< ::xercesc::DOMDocument const& > (*d), f);"
             << "}";


          os << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << id (name) << " (" << string_type << " const& u," << endl
             << "::xsd::cxx::xml::error_handler< " <<
            char_type << " >& h," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << properties_type << " const& p)"
             << "{"
             << "::xsd::cxx::xml::auto_initializer i (" << endl
             << "(f & ::xsd::cxx::tree::flags::dont_initialize) == 0," << endl
             << "(f & ::xsd::cxx::tree::flags::keep_dom) == 0);"
             << endl
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << "::xsd::cxx::xml::dom::dom< " << char_type << " > (" << endl
             << "u, h, p, (f & ::xsd::cxx::tree::flags::dont_validate) == 0));"
             << endl
             << "if (!d)"
             << "{"
             << "throw ::xsd::cxx::tree::parsing< " <<
            char_type << " > ();"
             << "}"
             << "return " << fq_name (e) << " (" << endl
             << "static_cast< ::xercesc::DOMDocument const& > (*d), f);"
             << "}";

          os << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << id (name) << " (" << string_type << " const& u," << endl
             << "::xercesc::DOMErrorHandler& h," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << properties_type << " const& p)"
             << "{"
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << "::xsd::cxx::xml::dom::dom< " << char_type << " > (" << endl
             << "u, h, p, (f & ::xsd::cxx::tree::flags::dont_validate) == 0));"
             << endl
             << "if (!d)"
             << "{"
             << "throw ::xsd::cxx::tree::parsing< " <<
            char_type << " > ();"
             << "}"
             << "return " << fq_name (e) << " (" << endl
             << "static_cast< ::xercesc::DOMDocument const& > (*d), f);"
             << "}";


          // istream
          //
          os << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << id (name) << " (::std::istream& is," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << properties_type << " const& p)"
             << "{"
             << "::xsd::cxx::xml::auto_initializer i (" << endl
             << "(f & ::xsd::cxx::tree::flags::dont_initialize) == 0," << endl
             << "(f & ::xsd::cxx::tree::flags::keep_dom) == 0);"
             << endl
             << "::xsd::cxx::xml::sax::std_input_source isrc (is);"
             << "xercesc::Wrapper4InputSource wrap (&isrc, false);"
             << "return " << fq_name (e) << " (wrap, f, p);"
             << "}";

          os << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << id (name) << " (::std::istream& is," << endl
             << "::xsd::cxx::xml::error_handler< " <<
            char_type << " >& h," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << properties_type << " const& p)"
             << "{"
             << "::xsd::cxx::xml::auto_initializer i (" << endl
             << "(f & ::xsd::cxx::tree::flags::dont_initialize) == 0," << endl
             << "(f & ::xsd::cxx::tree::flags::keep_dom) == 0);"
             << endl
             << "::xsd::cxx::xml::sax::std_input_source isrc (is);"
             << "xercesc::Wrapper4InputSource wrap (&isrc, false);"
             << "return " << fq_name (e) << " (wrap, h, f, p);"
             << "}";

          os << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << id (name) << " (::std::istream& is," << endl
             << "::xercesc::DOMErrorHandler& h," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << properties_type << " const& p)"
             << "{"
             << "::xsd::cxx::xml::sax::std_input_source isrc (is);"
             << "xercesc::Wrapper4InputSource wrap (&isrc, false);"
             << "return " << fq_name (e) << " (wrap, h, f, p);"
             << "}";

          os << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << id (name) << " (::std::istream& is," << endl
             << string_type << " const& sid," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << properties_type << " const& p)"
             << "{"
             << "::xsd::cxx::xml::auto_initializer i (" << endl
             << "(f & ::xsd::cxx::tree::flags::dont_initialize) == 0," << endl
             << "(f & ::xsd::cxx::tree::flags::keep_dom) == 0);"
             << endl
             << "::xsd::cxx::xml::sax::std_input_source isrc (is, sid);"
             << "xercesc::Wrapper4InputSource wrap (&isrc, false);"
             << "return " << fq_name (e) << " (wrap, f, p);"
             << "}";

          os << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << id (name) << " (::std::istream& is," << endl
             << string_type << " const& sid," << endl
             << "::xsd::cxx::xml::error_handler< " <<
            char_type << " >& h," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << properties_type << " const& p)"
             << "{"
             << "::xsd::cxx::xml::auto_initializer i (" << endl
             << "(f & ::xsd::cxx::tree::flags::dont_initialize) == 0," << endl
             << "(f & ::xsd::cxx::tree::flags::keep_dom) == 0);"
             << endl
             << "::xsd::cxx::xml::sax::std_input_source isrc (is, sid);"
             << "xercesc::Wrapper4InputSource wrap (&isrc, false);"
             << "return " << fq_name (e) << " (wrap, h, f, p);"
             << "}";

          os << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << id (name) << " (::std::istream& is," << endl
             << string_type << " const& sid," << endl
             << "::xercesc::DOMErrorHandler& h," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << properties_type << " const& p)"
             << "{"
             << "::xsd::cxx::xml::sax::std_input_source isrc (is, sid);"
             << "xercesc::Wrapper4InputSource wrap (&isrc, false);"
             << "return " << fq_name (e) << " (wrap, h, f, p);"
             << "}";


          // InputSource.
          //
          os << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << id (name) << " (::xercesc::DOMInputSource const& i," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << properties_type << " const& p)"
             << "{"
             << "::xsd::cxx::tree::error_handler< " << char_type << " > h;"
             << endl
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << "::xsd::cxx::xml::dom::dom< " << char_type << " > (" << endl
             << "i, h, p, (f & ::xsd::cxx::tree::flags::dont_validate) == 0));"
             << endl
             << "h.throw_if_failed< ::xsd::cxx::tree::parsing< " <<
            char_type << " > > ();"
             << endl
             << "return " << fq_name (e) << " (" << endl
             << "static_cast< ::xercesc::DOMDocument const& > (*d), f, p);"
             << "}";

          os << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << id (name) << " (::xercesc::DOMInputSource const& i," << endl
             << "::xsd::cxx::xml::error_handler< " <<
            char_type << " >& h," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << properties_type << " const& p)"
             << "{"
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << "::xsd::cxx::xml::dom::dom< " << char_type << " > (" << endl
             << "i, h, p, (f & ::xsd::cxx::tree::flags::dont_validate) == 0));"
             << endl
             << "if (!d)"
             << "{"
             << "throw ::xsd::cxx::tree::parsing< " <<
            char_type << " > ();"
             << "}"
             << "return " << fq_name (e) << " (" << endl
             << "static_cast< ::xercesc::DOMDocument const& > (*d), f, p);"
             << "}";


          os << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << id (name) << " (::xercesc::DOMInputSource const& i," << endl
             << "::xercesc::DOMErrorHandler& h," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << properties_type << " const& p)"
             << "{"
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << "::xsd::cxx::xml::dom::dom< " << char_type << " > (" << endl
             << "i, h, p, (f & ::xsd::cxx::tree::flags::dont_validate) == 0));"
             << endl
             << "if (!d)"
             << "{"
             << "throw ::xsd::cxx::tree::parsing< " <<
            char_type << " > ();"
             << "}"
             << "return " << fq_name (e) << " (" << endl
             << "static_cast< ::xercesc::DOMDocument const& > (*d), f, p);"
             << "}";


          // DOM.
          //

          Boolean fund (false);

          {
            IsFundamentalType t (fund);
            t.dispatch (e.type ());
          }

          // Check if we need to handle xsi:type and substitution groups.
          // If this element's type is anonymous or mapped to a fundamental
          // C++ type then we don't need to do anything.
          //
          Boolean polymorphic (e.type ().named () && !fund);

          os << "::std::auto_ptr< " << type_name (e) << " >" << endl
             << id (name) << " (::xercesc::DOMDocument const& d," << endl
             << "::xsd::cxx::tree::flags f," << endl
             << properties_type << " const&)"
             << "{"
             << "::xsd::cxx::xml::dom::auto_ptr< ::xercesc::DOMDocument > " <<
            "c (" << endl
             << "(f & ::xsd::cxx::tree::flags::keep_dom) " << endl
             << "? static_cast< ::xercesc::DOMDocument* > (" <<
            "d.cloneNode (true))" << endl
             << ": 0);"
             << endl
             << xml_element_type << " e (" << endl
             << "c.get () ? c->getDocumentElement () : d.getDocumentElement ());"
             << endl;

          if (polymorphic)
          {
            // aCC cannot handle an inline call to type_factory_map_instance.
            //
            os << "::xsd::cxx::tree::type_factory_map< " << char_type <<
              " >& tfm (" << endl
               << "::xsd::cxx::tree::type_factory_map_instance< 0, " <<
              char_type << " > ());"
               << endl
               << "::std::auto_ptr< " << type_name (e) << " > r (" << endl
               << "tfm.create< " << type_name (e) << " > (" << endl
               << L << "\"" << name << "\"," << endl
               << L << "\"" << e.namespace_().name () << "\"," << endl
               << "true," << endl
               << "true," << endl
               << "e," << endl
               << "f," << endl
               << "0));"
               << endl
               << "if (r.get () != 0)"
               << "{";
          }
          else
          {
            os << "if (e.name () == " << L << "\"" << name << "\" && " <<
              "e.namespace_ () == " << L << "\"" << e.namespace_().name () << "\")"
               << "{";

            if (fund)
            {
              os << "::std::auto_ptr< " << type_name (e) << " > r (" <<
                "new " << type_name (e) << " (" <<
                "::xsd::cxx::tree::traits< " << type_name (e) << " >::" <<
                "create (e, f, 0)));";
            }
            else
            {
              os << "::std::auto_ptr< " << type_name (e) << " > r (" <<
                "::xsd::cxx::tree::traits< " << type_name (e) << " >::" <<
                "create (e, f, 0));";
            }
          }

          os << "if (f & ::xsd::cxx::tree::flags::keep_dom) c.release ();"
             << "return r;"
             << "}"
             << "throw ::xsd::cxx::tree::unexpected_element < " <<
            char_type << " > (" << endl
             << "e.name ()," << endl
             << "e.namespace_ ()," << endl
             << L << "\"" << name << "\"," << endl
             << L << "\"" << e.namespace_().name () << "\");"
             << "}";
        }

      private:
        String
        type_name (Type& e)
        {
          SemanticGraph::Type& t (e.type ());

          if (t.named ())
          {
            std::wostringstream o;

            MemberTypeName type (*this, o);
            type.dispatch (t);

            return o.str ();
          }
          else
          {
            String name (id (e.name ()));

            return name + L"::" + (name == L"type" ? L"type_" : L"type");
          }
        }
      };
    }

    Void
    generate_parser_source (Context& ctx, SemanticGraph::Schema& tu)
    {
      ctx.os << "#include <istream>" << endl
             << "#include <xercesc/framework/Wrapper4InputSource.hpp>" << endl
             << "#include <xsd/cxx/xml/sax/std-input-source.hxx>" << endl
             << "#include <xsd/cxx/tree/exceptions.hxx>" << endl
             << "#include <xsd/cxx/tree/error-handler.hxx>" << endl
             << endl;

      Traversal::Schema schema;
      Traversal::Sources sources;
      Traversal::Names names_ns, names;
      Namespace ns (ctx);
      Element element (ctx);

      schema >> sources >> schema;
      schema >> names_ns >> ns >> names >> element;

      schema.dispatch (tu);
    }
  }
}
