// file      : xsd/cxx/tree/serialization-source.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/serialization-source.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      struct List : Traversal::List, protected virtual Context
      {
        List (Context& c,
              String const& name_ = "",
              String const& type_name_ = "")
            : Context (c),
              name (name_),
              type_name (type_name_)
        {
        }

        virtual Void
        traverse (Type& l)
        {
          Boolean named (l.named ());

          if (named)
          {
            name = l.name ();

            enter_scope (id (name));
            type_name = scope;
            leave_scope ();
          }


          // operator<< (xml::element)
          //
          os << "void" << endl
             << "operator<< (" << xml_element_type << "& e," << endl
             << (named ? "class " : "") << type_name << " const& v)"
             << "{"
             << "e << static_cast< ::xsd::cxx::tree::sequence< " <<
            item_type_name (l.argumented ().type ()) << " > const& > (v);"
             << "}";

          // operator<< (xml::attribute)
          //
          os << "void" << endl
             << "operator<< (" << xml_attribute_type << "& a," << endl
             << (named ? "class " : "") << type_name << " const& v)"
             << "{"
             << "a << static_cast< ::xsd::cxx::tree::sequence< " <<
            item_type_name (l.argumented ().type ()) << " > const& > (v);"
             << "}";

          // operator<< (list_stream)
          //
          os << "void" << endl
             << "operator<< (::xsd::cxx::xml::dom::list_stream< " << char_type
             << " >& l," << endl
             << (named ? "class " : "") << type_name << " const& v)"
             << "{"
             << "l << static_cast< ::xsd::cxx::tree::sequence< " <<
            item_type_name (l.argumented ().type ()) << " > const& > (v);"
             << "}";

          // Register with type factory map.
          //
          if (l.named ())
          {
            os << "static" << endl
               << "::xsd::cxx::tree::type_name_initializer< 0, " <<
              char_type << ", " <<
              (named ? "class " : "") << type_name << " >" << endl
               << id ("_xsd_", name, "_type_name_init") << " (" << endl
               << L << "\"" << name << "\"," << endl
               << L << "\"" << xs_ns_name (l) << "\");"
               << endl
               << endl;
          }
        }

      private:
        String
        item_type_name (SemanticGraph::Type& t)
        {
          // This type is always named.
          //
          std::wostringstream o;

          MemberTypeName type (*this, o);
          type.dispatch (t);

          return o.str ();
        }

        /*
        String
        type_name (Type& l)
        {
          if (l.named ())
          {
            std::wostringstream o;

            MemberTypeName type (*this, o);
            type.dispatch (l);

            return o.str ();
          }
          else
          {
            enter_class_scope (id (name));
            String scoped_name (scope);
            leave_scope ();

            return scoped_name;
          }
        }
        */

      private:
        String name;
        String type_name;
      };


      struct Union : Traversal::Union, protected virtual Context
      {
        Union (Context& c,
               String const& name_ = "",
               String const& type_name_ = "")
            : Context (c),
              name (name_),
              type_name (type_name_)
        {
        }

        virtual Void
        traverse (Type& u)
        {
          Boolean named (u.named ());

          if (named)
          {
            name = u.name ();

            enter_scope (id (name));
            type_name = scope;
            leave_scope ();
          }

          String base (L"::xsd::cxx::tree::string< " + char_type + L" >");

          // operator<< (xml::element)
          //
          os << "void" << endl
             << "operator<< (" << xml_element_type << "& e," << endl
             << (named ? "class " : "") << type_name << " const& v)"
             << "{"
             << "e << static_cast< " << base << " const& > (v);"
             << "}";

          // operator<< (xml::attribute)
          //
          os << "void" << endl
             << "operator<< (" << xml_attribute_type << "& a," << endl
             << (named ? "class " : "") << type_name << " const& v)"
             << "{"
             << "a << static_cast< " << base << " const& > (v);"
             << "}";

          // operator<< (list_stream)
          //
          os << "void" << endl
             << "operator<< (::xsd::cxx::xml::dom::list_stream< " << char_type
             << " >& l," << endl
             << (named ? "class " : "") << type_name << " const& v)"
             << "{"
             << "l << static_cast< " << base << " const& > (v);"
             << "}";

          // Register with type factory map.
          //
          if (u.named ())
          {
            os << "static" << endl
               << "::xsd::cxx::tree::type_name_initializer< 0, " <<
              char_type << ", " <<
              (named ? "class " : "") << type_name << " >" << endl
               << id ("_xsd_", name, "_type_name_init") << " (" << endl
               << L << "\"" << name << "\"," << endl
               << L << "\"" << xs_ns_name (u) << "\");"
               << endl
               << endl;
          }
        }

      private:
        /*
        String
        type_name (Type& u)
        {
          if (u.named ())
          {
            std::wostringstream o;

            MemberTypeName type (*this, o);
            type.dispatch (u);

            return o.str ();
          }
          else
          {
            enter_class_scope (id (name));
            String scoped_name (scope);
            leave_scope ();

            return scoped_name;
          }
        }
        */

      private:
        String name;
        String type_name;
      };


      struct EnumeratorToMember : Traversal::Enumerator,
                                  protected virtual Context
      {
        EnumeratorToMember (Context& c, String const& scope_)
            : Context (c),
              scope (scope_)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          String name (e.name ());

          os << "case " << scope << "::" << id (name) << ":"
             << "{"
             << "n.value (" << L << "\"" << name << "\");"
             << "break;"
             << "}";
        }

      private:
        String scope;
      };


      struct EnumeratorToStream : Traversal::Enumerator,
                                  protected virtual Context
      {
        EnumeratorToStream (Context& c, String const& scope_)
            : Context (c),
              scope (scope_)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          String name (e.name ());

          os << "case " << scope << "::" << id (name) << ":"
             << "{"
             << "l << " << L << "\"" << name << "\";"
             << "break;"
             << "}";
        }

      private:
        String scope;
      };


      struct Enumeration : Traversal::Enumeration, protected virtual Context
      {
        Enumeration (Context& c,
                     String const& name_ = "",
                     String const& type_name_ = "")
            : Context (c),
              name (name_),
              type_name (type_name_)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          Boolean named (e.named ());

          if (named)
          {
            name = e.name ();

            enter_class_scope (id (name), e);
            type_name = scope;
            leave_scope ();
          }

          Traversal::Names names_to_member;
          EnumeratorToMember enumerator_to_member (*this, type_name);

          names_to_member >> enumerator_to_member;

          Traversal::Names names_to_stream;
          EnumeratorToStream enumerator_to_stream (*this, type_name);

          names_to_stream >> enumerator_to_stream;


          // operator<< (xml::element)
          //
          os << "void" << endl
             << "operator<< (" << xml_element_type << "& n," << endl
             << (named ? "class " : "") << type_name << " i)"
             << "{"
             << "switch (i)"
             << "{";

          {
            enter_class_scope (id (name), e);
            names (e, names_to_member);
            leave_scope ();
          }

          os << "}" // switch
             << "}";


          // operator<< (xml::attribute)
          //
          os << "void" << endl
             << "operator<< (" << xml_attribute_type << "& n," << endl
             << (named ? "class " : "") << type_name << " i)"
             << "{"
             << "switch (i)"
             << "{";

          {
            enter_class_scope (id (name), e);
            names (e, names_to_member);
            leave_scope ();
          }

          os << "}" // switch
             << "}";

          // operator<< (list_stream)
          //
          os << "void" << endl
             << "operator<< (::xsd::cxx::xml::dom::list_stream< " << char_type
             << " >& l," << endl
             << (named ? "class " : "") << type_name << " i)"
             << "{"
             << "switch (i)"
             << "{";

          {
            enter_class_scope (id (name), e);
            names (e, names_to_stream);
            leave_scope ();
          }

          os << "}" // switch
             << "}";


          // Register with type factory map.
          //
          if (e.named ())
          {
            os << "static" << endl
               << "::xsd::cxx::tree::type_name_initializer< 0, " <<
              char_type << ", " <<
              (named ? "class " : "") << type_name << " >" << endl
               << id ("_xsd_", name, "_type_name_init") << " (" << endl
               << L << "\"" << name << "\"," << endl
               << L << "\"" << xs_ns_name (e) << "\");"
               << endl
               << endl;
          }
        }

      private:
        /*
        String
        type_name (Type& e)
        {
          if (e.named ())
          {
            std::wostringstream o;

            MemberTypeName type (*this, o);
            type.dispatch (e);

            return o.str ();
          }
          else
          {
            enter_class_scope (id (name));
            String scoped_name (scope);
            leave_scope ();

            return scoped_name;
          }
        }
        */

      private:
        String name;
        String type_name;
      };


      //
      //
      template <typename Complex>
      struct Member : Traversal::Member, protected virtual Context
      {
        Member (Context& c, String const& scope_)
            : Context (c),
              scope (scope_)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          SemanticGraph::Type& t (m.type ());

          if (!t.named () && !t.context ().count ("seen"))
          {
            // I need to build scope information with the same names
            // as in the header so that I get member escaping right.
            // Actual names will be from the public interface.
            //

            String name (m.name ());
            String escaped_name (id (name));
            String escaped_xsd_name (id ("_xsd_", name));
            String escaped_xsd_name_ (id ("_xsd_", name, "_"));

            String type_name (scope
                              + L"::" + escaped_name
                              + L"::" + unclash (name, "type"));

            Traversal::Belongs belongs_;

            List list (*this, escaped_name, type_name);
            Union union_ (*this, escaped_name, type_name);
            Complex complex (*this, escaped_name, type_name);
            Enumeration enumeration (*this, escaped_name, type_name);

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            enter_scope (escaped_xsd_name);
            enter_scope (escaped_xsd_name_);

            t.context ().set ("seen", true);

            belongs (m, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }
        }

      private:
        String scope;
      };


      struct Element : Traversal::Element, protected virtual Context
      {
        Element (Context& c, String const& scope_)
            : Context (c),
              scope (scope_)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          String name (e.name ());
          String ns (e.qualified () ? e.namespace_ ().name () : "");
          String type (scope
                       + L"::" + id (name)
                       + L"::" + unclash (name, "type"));

          // Check if we need to handle xsi:type and substitution groups.
          // If this element's type is anonymous or mapped to a fundamental
          // C++ type then we don't need to do anything.
          //
          SemanticGraph::Type& t (e.type ());
          Boolean polymorphic (t.named ());

          if (polymorphic)
          {
            Boolean fund (false);
            IsFundamentalType traverser (fund);
            traverser.dispatch (t);
            polymorphic = !fund;
          }

          os << "{";

          // aCC cannot handle an inline call to type_name_map_instance.
          //
          if (polymorphic)
            os << "::xsd::cxx::tree::type_name_map< " << char_type
               << " >& tnm (" << endl
               << "::xsd::cxx::tree::type_name_map_instance< 0, " <<
              char_type << " > ());"
               << endl;

          if (e.max () != 1)
          {
            // sequence
            //
            os << "for (" << scope << "::" << id (name) << "::" <<
              unclash (name, "const_iterator") << endl
               << "b (i." << id (name) << " ().begin ()), " <<
              "e (i." << id (name) << " ().end ());" << endl
               << "b != e; ++b)"
               << "{";

            if (polymorphic)
            {
              os << "tnm.serialize< " << type << " > (" << endl
                 << L << "\"" << name << "\"," << endl
                 << L << "\"" << ns << "\"," << endl
                 << (e.global () ? "true" : "false") << "," << endl
                 << (e.qualified () ? "true" : "false") << "," << endl
                 << "n," << endl
                 << "*b);";
            }
            else
            {
              os << xml_element_type << " s (" << endl
                 << L << '"' << name << "\"," << endl
                 << (ns ? L + L'"' + ns + L'"' + L",\n" : L"")
                 << "n);"
                 << "s << *b;";
            }

            os << "}";
          }
          else if (e.min () == 0)
          {
            // optional
            //
            os << "if (i." << id (name) << " ())"
               << "{";

            if (polymorphic)
            {
              os << "tnm.serialize< " << type << " > (" << endl
                 << L << "\"" << name << "\"," << endl
                 << L << "\"" << ns << "\"," << endl
                 << (e.global () ? "true" : "false") << "," << endl
                 << (e.qualified () ? "true" : "false") << "," << endl
                 << "n," << endl
                 << "*i." << id (name) << " ());";
            }
            else
            {
              os << xml_element_type << " e (" << endl
                 << L << '"' << name << "\"," << endl
                 << (ns ? L + L'"' + ns + L'"' + L",\n" : L"")
                 << "n);"
                 << "e << *i." << id (name) << " ();";
            }

            os << "}";
          }
          else
          {
            // one
            //
            if (polymorphic)
            {
              os << "tnm.serialize< " << type << " > (" << endl
                 << L << "\"" << name << "\"," << endl
                 << L << "\"" << ns << "\"," << endl
                 << (e.global () ? "true" : "false") << "," << endl
                 << (e.qualified () ? "true" : "false") << "," << endl
                 << "n," << endl
                 << "i." << id (name) << " ());";
            }
            else
            {
              os << xml_element_type << " e (" << endl
                 << L << '"' << name << "\"," << endl
                 << (ns ? L + L'"' + ns + L'"' + L",\n" : L"")
                 << "n);"
                 << "e << i." << id (name) << " ();";
            }
          }

          os << "}";
        }

      private:
        String scope;
      };

      struct Attribute : Traversal::Attribute, protected virtual Context
      {
        Attribute (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& a)
        {
          String name (a.name ());
          String ns (a.qualified () ? a.namespace_ ().name () : "");

          if (a.optional () && !a.default_ ())
          {
            os << "if (i." << id (name) << " ())"
               << "{"
               << xml_attribute_type << " a (" << endl
               << L << '"' << name << "\"," << endl
               << (ns ? L + L'"' + ns + L'"' + L",\n" : L"")
               << "n);"
               << endl
               << "a << *i." << id (name) << " ();"
               << "}";
          }
          else
          {
            os << "{"
               << xml_attribute_type << " a (" << endl
               << L << '"' << name << "\"," << endl
               << (ns ? L + L'"' + ns + L'"' + L",\n" : L"")
               << "n);"
               << endl
               << "a << i." << id (name) << " ();"
               << "}";
          }
        }
      };


      struct Complex : Traversal::Complex, protected virtual Context
      {
        Complex (Context& c,
                 String const& name_ = "",
                 String const& type_name_ = L"")
            : Context (c),
              name (name_),
              type_name (type_name_),
              base_ (c)
        {
          inherits_ >> base_;
        }

        virtual Void
        traverse (Type& c)
        {
          Boolean named (c.named ());

          if (named)
          {
            name = c.name ();

            enter_class_scope (id (name), c);
            type_name = scope;
            leave_scope ();
          }

          // Go after anonymous types first.
          //
          {
            enter_class_scope (id (name), c);

            Traversal::Names names_member;
            Member<Complex> member (*this, type_name);

            names_member >> member;

            names (c, names_member);

            leave_scope ();
          }


          // operator<< (xml::element)
          //
          os << "void" << endl
             << "operator<< (" << xml_element_type << "& n," << endl
             << (named ? "class " : "") << type_name << " const& i)"
             << "{";

          inherits (c,
                    inherits_,
                    &Complex::inherits_pre,
                    &Complex::inherits_post,
                    &Complex::inherits_none);


          {
            enter_class_scope (id (name), c);

            Traversal::Names names_member;
            Element element (*this, type_name);
            Attribute attribute (*this);

            names_member >> element;
            names_member >> attribute;

            names (c, names_member);

            leave_scope ();
          }

          os << "}";


          //@@ Need to factor this out into serialization-elements
          //
          Boolean simple (true);

          struct Traverser : Traversal::Complex
          {
            Traverser (Boolean& v)
                : v_ (v)
            {
              edge_traverser (inherits_);
              inherits_.node_traverser (*this);
            }

            virtual Void
            traverse (SemanticGraph::Complex& c)
            {
              if (has<Traversal::Element> (c) ||
                  has<Traversal::Attribute> (c))
                v_ = false;
              else
                inherits (c, *this);
            }

          private:
            Traversal::Inherits inherits_;
            Boolean& v_;
          } t (simple);

          t.dispatch (c);

          if (simple)
          {
            // operator<< (xml::attribute)
            //
            os << "void" << endl
               << "operator<< (" << xml_attribute_type << "& n," << endl
               << (named ? "class " : "") << type_name << " const& i)"
               << "{";

            inherits (c,
                      inherits_,
                      &Complex::inherits_pre,
                      &Complex::inherits_post,
                      &Complex::inherits_none);

            os << "}";


            // operator<< (list_stream)
            //
            os << "void" << endl
               << "operator<< (::xsd::cxx::xml::dom::list_stream< " << char_type
               << " >& n," << endl
               << (named ? "class " : "") << type_name << " const& i)"
               << "{";

            inherits (c,
                      inherits_,
                      &Complex::inherits_pre,
                      &Complex::inherits_post,
                      &Complex::inherits_none);

            os << "}";
          }

          // Register with type factory map.
          //
          if (c.named ())
          {
            os << "static" << endl
               << "::xsd::cxx::tree::type_name_initializer< 0, " <<
              char_type << ", " <<
              (named ? "class " : "") << type_name << " >" << endl
               << id ("_xsd_", name, "_type_name_init") << " (" << endl
               << L << "\"" << name << "\"," << endl
               << L << "\"" << xs_ns_name (c) << "\");"
               << endl
               << endl;
          }
        }

        virtual Void
        inherits_pre (Type& t)
        {
          os << "n << static_cast< ";
        }

        virtual Void
        inherits_post (Type& t)
        {
          os << " const& > (i);";
        }

        virtual Void
        inherits_none (Type& t)
        {
        }

      private:
        /*
        String
        type_name (Type& c)
        {
          if (c.named ())
          {
            std::wostringstream o;

            MemberTypeName type (*this, o);
            type.dispatch (c);

            return o.str ();
          }
          else
          {
            enter_class_scope (id (name));
            String scoped_name (scope);
            leave_scope ();

            return scoped_name;
          }
        }
        */

      private:
        String name;
        String type_name;

        Traversal::Inherits inherits_;
        BaseTypeName base_;
      };


      // Goes after anonymous types via global element. Note that we don't
      // need to go after attributes because they can only be referenced
      // and in that case everything will be generated at the point of
      // reference.
      //
      struct GlobalMember : Traversal::Element, protected virtual Context
      {
        GlobalMember (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          SemanticGraph::Type& t (e.type ());

          if (!t.named ())
          {
            String name (e.name ());
            String escaped_name (id (name));
            String escaped_xsd_name_ (id ("_xsd_", name, "_"));

            String type_name (escaped_name + L"::" + unclash (name, "type"));

            Traversal::Belongs belongs_;

            List list (*this, escaped_name, type_name);
            Union union_ (*this, escaped_name, type_name);
            Complex complex (*this, escaped_name, type_name);
            Enumeration enumeration (*this, escaped_name, type_name);

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            enter_scope (escaped_name);
            enter_scope (escaped_xsd_name_);

            t.context ().set ("seen", true);

            belongs (e, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }

          if (SemanticGraph::Substitutes* s = e.substitutes ())
          {
            Type& r (s->root ());

            String type_name;

            if (t.named ())
            {
              std::wostringstream o;

              MemberTypeName type (*this, o);
              type.dispatch (t);

              type_name = o.str ();
            }
            else
            {
              String name (id (e.name ()));
              type_name = name + L"::" +
                (name == L"type" ? L"type_" : L"type");
            }

            os << "static" << endl
               << "::xsd::cxx::tree::type_name_initializer< 0, " <<
              char_type << ", " << type_name << " >" << endl
               << id ("_xsd_", e.name (), "_element_name_init") << " (" << endl
               << L << "\"" << r.name () << "\"," << endl
               << L << "\"" << r.namespace_ ().name () << "\"," << endl
               << L << "\"" << e.name () << "\"," << endl
               << L << "\"" << e.namespace_ ().name () << "\");"
               << endl
               << endl;
          }
        }
      };


      struct Serializer: Traversal::Element,
                         protected virtual Context
      {
        Serializer (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          String name (e.name ());
          String ns (e.namespace_ ().name ());

          // Note that I am using fq-name in function calls because g++ gets
          // confused if the name is 'type'. (see tests/schema/anonymous)
          //

          Boolean fund (false);
          {
            IsFundamentalType t (fund);
            t.dispatch (e.type ());
          }

          // Check if we need to handle xsi:type and substitution groups.
          // If this element's type is anonymous or mapped to a fundamental
          // C++ type then we don't need to do anything.
          //
          Boolean polymorphic (e.type ().named () && !fund);


          // To an existing DOM instance.
          //
          os << "void" << endl
             << id (name) << " (::xercesc::DOMDocument& d," << endl
             << type_name (e) << " const& s," << endl
             << "::xsd::cxx::tree::flags)"
             << "{"
             << xml_element_type << " e (d.getDocumentElement ());"
             << endl;

          if (polymorphic)
          {
            // aCC cannot handle an inline call to type_name_map_instance.
            //
            os << "::xsd::cxx::tree::type_name_map< " << char_type
               << " >& tnm (" << endl
               << "::xsd::cxx::tree::type_name_map_instance< 0, " <<
              char_type << " > ());"
               << endl
               << "tnm.serialize< " << type_name (e) << " > (" << endl
               << L << "\"" << name << "\"," << endl
               << L << "\"" << e.namespace_().name () << "\"," << endl
               << "e," << endl
               << "s);";
          }
          else
          {
            os << "if (e.name () == " << L << "\"" << name << "\" && " <<
              "e.namespace_ () == " << L << "\"" << ns << "\")"
               << "{"
               << "e << s;"
               << "}"
               << "else"
               << "{"
               << "throw ::xsd::cxx::tree::unexpected_element < " <<
              char_type << " > (e.name (), e.namespace_ (), " <<
              L << "\"" << name << "\", " << L << "\"" << ns << "\");"
               << "}";
          }

          os << "}";


          // To a new DOM instance.
          //
          os << "::xsd::cxx::xml::dom::auto_ptr< ::xercesc::DOMDocument >" << endl
             << id (name) << " (" << type_name (e) << " const& s," << endl
             << "::xsd::cxx::xml::dom::namespace_infomap< " << char_type <<
            "> const& m," << endl
             << "::xsd::cxx::tree::flags)"
             << "{"
             << "try"
             << "{";


          if (polymorphic)
          {
            // aCC cannot handle an inline call to type_name_map_instance.
            //
            os << "::xsd::cxx::tree::type_name_map< " << char_type
               << " >& tnm (" << endl
               << "::xsd::cxx::tree::type_name_map_instance< 0, " <<
              char_type << " > ());"
               << endl;
          }

          os << "::xsd::cxx::xml::dom::auto_ptr< ::xercesc::DOMDocument > d (" << endl;

          if (polymorphic)
          {
            os << "tnm.dom< " << type_name (e) << " > (" << endl
               << L << "\"" << name << "\"," << endl
               << L << "\"" << e.namespace_().name () << "\"," << endl
               << "m," << endl
               << "s));";
          }
          else
          {
            os << "::xsd::cxx::xml::dom::dom< " << char_type << " > (" << endl
               << L << "\"" << name << "\"," << endl
               << L << "\"" << ns << "\", m));";
          }

          os << fq_name (e) << " (*d, s);"
             << "return d;"
             << "}"
             << "catch (::xsd::cxx::xml::dom::mapping< " << char_type << " > const& e)"
             << "{"
             << "throw ::xsd::cxx::tree::no_namespace_mapping< " << char_type << " > (e.name ());"
             << "}"
             << "catch (::xsd::cxx::xml::dom::xsi_already_in_use const&)"
             << "{"
             << "throw ::xsd::cxx::tree::xsi_already_in_use< " <<
	     char_type << " > ();"
             << "}"
             << "}";


          // To XMLFormatTarget.
          //
          os << "void" << endl
             << id (name) << " (::xercesc::XMLFormatTarget& t," << endl
             << type_name (e) << " const& s," << endl
             << "::xsd::cxx::xml::dom::namespace_infomap< " << char_type <<
            "> const& m," << endl
             << "::std::basic_string< " << char_type << " > const& e," << endl
             << "::xsd::cxx::tree::flags)"
             << "{"
             << "::xsd::cxx::xml::dom::auto_ptr< ::xercesc::DOMDocument > d (" <<
            fq_name (e) << " (s, m));"
             << endl
             << "::xsd::cxx::tree::error_handler< " << char_type << " > h;"
             << endl
             << "if (!::xsd::cxx::xml::dom::serialize (t, *d, e, h))"
             << "{"
             << "h.throw_if_failed< ::xsd::cxx::tree::serialization< " <<
            char_type << " > > ();"
             << "}"
             << "}";

          os << "void" << endl
             << id (name) << " (::xercesc::XMLFormatTarget& t," << endl
             << type_name (e) << " const& s," << endl
             << "::xsd::cxx::xml::dom::namespace_infomap< " << char_type <<
            "> const& m," << endl
             << "::xsd::cxx::xml::error_handler< " <<
            char_type << " >& h," << endl
             << "::std::basic_string< " << char_type << " > const& e," << endl
             << "::xsd::cxx::tree::flags)"
             << "{"
             << "::xsd::cxx::xml::dom::auto_ptr< ::xercesc::DOMDocument > d (" <<
            fq_name (e) << " (s, m));"
             << "if (!::xsd::cxx::xml::dom::serialize (t, *d, e, h))"
             << "{"
             << "throw ::xsd::cxx::tree::serialization< " <<
            char_type << " > ();"
             << "}"
             << "}";

          os << "void" << endl
             << id (name) << " (::xercesc::XMLFormatTarget& t," << endl
             << type_name (e) << " const& s," << endl
             << "::xsd::cxx::xml::dom::namespace_infomap< " << char_type <<
            "> const& m," << endl
             << "::xercesc::DOMErrorHandler& h," << endl
             << "::std::basic_string< " << char_type << " > const& e," << endl
             << "::xsd::cxx::tree::flags)"
             << "{"
             << "::xsd::cxx::xml::dom::auto_ptr< ::xercesc::DOMDocument > d (" <<
            fq_name (e) << " (s, m));"
             << "if (!::xsd::cxx::xml::dom::serialize (t, *d, e, h))"
             << "{"
             << "throw ::xsd::cxx::tree::serialization< " <<
            char_type << " > ();"
             << "}"
             << "}";


          // To std::ostream.
          //
          os << "void" << endl
             << id (name) << " (::std::ostream& o," << endl
             << type_name (e) << " const& s," << endl
             << "::xsd::cxx::xml::dom::namespace_infomap< " << char_type <<
            "> const& m," << endl
             << "::std::basic_string< " << char_type << " > const& e," << endl
             << "::xsd::cxx::tree::flags f)"
             << "{"
             << "::xsd::cxx::xml::auto_initializer i (" << endl
             << "(f & ::xsd::cxx::tree::flags::dont_initialize) == 0);"
             << endl
             << "::xsd::cxx::xml::dom::auto_ptr< ::xercesc::DOMDocument > d (" <<
            fq_name (e) << " (s, m));"
             << endl
             << "::xsd::cxx::tree::error_handler< " << char_type << " > h;"
             << endl
             << "::xsd::cxx::xml::dom::ostream_format_target t (o);"
             << "if (!::xsd::cxx::xml::dom::serialize (t, *d, e, h))"
             << "{"
             << "h.throw_if_failed< ::xsd::cxx::tree::serialization< " <<
            char_type << " > > ();"
             << "}"
             << "}";

          os << "void" << endl
             << id (name) << " (::std::ostream& o," << endl
             << type_name (e) << " const& s," << endl
             << "::xsd::cxx::xml::dom::namespace_infomap< " << char_type <<
            "> const& m," << endl
             << "::xsd::cxx::xml::error_handler< " <<
            char_type << " >& h," << endl
             << "::std::basic_string< " << char_type << " > const& e," << endl
             << "::xsd::cxx::tree::flags f)"
             << "{"
             << "::xsd::cxx::xml::auto_initializer i (" << endl
             << "(f & ::xsd::cxx::tree::flags::dont_initialize) == 0);"
             << endl
             << "::xsd::cxx::xml::dom::auto_ptr< ::xercesc::DOMDocument > d (" <<
            fq_name (e) << " (s, m));"
             << "::xsd::cxx::xml::dom::ostream_format_target t (o);"
             << "if (!::xsd::cxx::xml::dom::serialize (t, *d, e, h))"
             << "{"
             << "throw ::xsd::cxx::tree::serialization< " <<
            char_type << " > ();"
             << "}"
             << "}";

          os << "void" << endl
             << id (name) << " (::std::ostream& o," << endl
             << type_name (e) << " const& s," << endl
             << "::xsd::cxx::xml::dom::namespace_infomap< " << char_type <<
            "> const& m," << endl
             << "::xercesc::DOMErrorHandler& h," << endl
             << "::std::basic_string< " << char_type << " > const& e," << endl
             << "::xsd::cxx::tree::flags)"
             << "{"
             << "::xsd::cxx::xml::dom::auto_ptr< ::xercesc::DOMDocument > d (" <<
            fq_name (e) << " (s, m));"
             << "::xsd::cxx::xml::dom::ostream_format_target t (o);"
             << "if (!::xsd::cxx::xml::dom::serialize (t, *d, e, h))"
             << "{"
             << "throw ::xsd::cxx::tree::serialization< " <<
            char_type << " > ();"
             << "}"
             << "}";

        }

      private:
        String
        type_name (Type& e)
        {
          SemanticGraph::Type& t (e.type ());

          if (t.named ())
          {
            /*
            // @@ g++-3.3 can handle "class ::ns::name" but happily
            //    accepts "class name".
            //
            if (t.name () == e.name () &&
                t.scope ().name () == e.scope ().name ()) // same namespace
            {
              // Special case for g++-3.3
              //
              return L"class " + id (t.name ());
            }
            else
            {
            */

            std::wostringstream o;

            MemberTypeName type (*this, o);
            type.dispatch (t);

            return o.str ();
          }
          else
          {
            String name (id (e.name ()));

            return name + L"::" + unclash (name, "type");
          }
        }
      };
    }

    Void
    generate_serialization_source (Context& ctx, SemanticGraph::Schema& tu)
    {
      ctx.os << "#include <xsd/cxx/tree/type-name-map.hxx>" << endl
             << endl;

      ctx.os << "namespace _xsd"
             << "{"
             << "static" << endl
             << "::xsd::cxx::tree::type_name_plate< 0, " <<
        ctx.char_type << " >" << endl
             << "type_name_plate_init;"
             << "}";

      Traversal::Schema schema;

      Traversal::Sources sources;
      Traversal::Names names_ns, names;

      Namespace ns (ctx);

      List list (ctx);
      Union union_ (ctx);
      Complex complex (ctx);
      Enumeration enumeration (ctx);
      GlobalMember member (ctx);
      Serializer serializer (ctx);


      schema >> sources >> schema;
      schema >> names_ns >> ns >> names;

      names >> list;
      names >> union_;
      names >> complex;
      names >> enumeration;
      names >> member;
      names >> serializer;


      schema.dispatch (tu);
    }
  }
}
